import argparse
import getpass
import logging
import os
import re
import sys

import keyring
from rich.console import Console
from rich.logging import RichHandler
from rich.prompt import Prompt
from rich.table import Table

from .AiRegistry import AiRegistry
from .keprompt_functions import DefinedToolsArray
from .keprompt_vm import VM, print_prompt_code
from .version import __version__

console = Console()
console.size = console.size

logging.getLogger().setLevel(logging.WARNING)

FORMAT = "%(message)s"
# logging.basicConfig(level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler(console=console)])

logging.basicConfig(level=logging.WARNING,  format=FORMAT,datefmt="[%X]",handlers=[RichHandler(console=console, rich_tracebacks=True)])
log = logging.getLogger(__file__)


def print_functions():
    table = Table(title="Available Functions")
    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("Description/Parameters", style="green")
    # Sort by LLM name, then model.
    sortable_keys = [f"{AiRegistry.models[model_name].company}:{model_name}" for model_name in AiRegistry.models.keys()]
    sortable_keys.sort()

    for tool in DefinedToolsArray:
        function = tool['function']
        name = function['name']
        description = function['description']

        table.add_row(name, description,)
        for k,v in function['parameters']['properties'].items():
            table.add_row("", f"[bold blue]{k:10}[/]: {v['description']}")

        table.add_row("","")

    console.print(table)

def print_models():
    table = Table(title="Available Models")
    table.add_column("Company", style="cyan", no_wrap=True)
    table.add_column("Model", style="green")
    table.add_column("Max Token", style="magenta", justify="right")
    table.add_column("$/mT In", style="green", justify="right")
    table.add_column("$/mT Out", style="green", justify="right")

    # Sort by LLM name, then model.
    sortable_keys = [f"{AiRegistry.models[model_name].company}:{model_name}" for model_name in AiRegistry.models.keys()]
    sortable_keys.sort()

    last_company = ''
    for k in sortable_keys:
        company, model_name = k.split(':', maxsplit=1)
        model = AiRegistry.get_model(model_name)
        if company != last_company:
            table.add_row(company,model_name,str(model.context),f"{model.input*1_000_000:06.4f}",f"{model.output*1_000_000:06.4f}")
            last_company = company
        else:
            table.add_row("", model_name, str(model.context), f"{model.input*1_000_000:06.4f}",f"{model.output*1_000_000:06.4f}")

    console.print(table)

def print_prompt_names(prompt_files: list[str]) -> None:

    table = Table(title="Prompt Files")
    table.add_column("Prompt", style="cyan", no_wrap=True)
    table.add_column("Description", style="magenta")

    for prompt_file in prompt_files:
        try:
            with open(prompt_file, 'r') as file:
                first_line = file.readline().strip()[2:]  # Read first line
        except Exception as e:
            first_line = f"Error reading file: {str(e)}"

        table.add_row(os.path.basename(prompt_file), first_line)

    console.print(table)

def create_dropdown(options: list[str], prompt_text: str = "Select an option") -> str:
    # Display numbered options
    for i, option in enumerate(options, 1):
        console.print(f"{i}. {option}", style="cyan")

    # Get user input with validation
    while True:
        choice = Prompt.ask(
            prompt_text,
            choices=[str(i) for i in range(1, len(options) + 1)],
            show_choices=False
        )

        return options[int(choice) - 1]

def get_new_api_key() -> None:

    companies = sorted(AiRegistry.handlers.keys())
    company = create_dropdown(companies, "AI Company?")
    # api_key = console.input(f"[bold green]Please enter your [/][bold cyan]{company} API key: [/]")
    api_key = getpass.getpass(f"Please enter your {company} API key: ")
    keyring.set_password("keprompt", username=company, password=api_key)

def print_prompt_lines(prompts_files: list[str]) -> None:
    table = Table(title="Prompt Code")
    table.add_column("Prompt", style="cyan bold", no_wrap=True)
    table.add_column("Lno", style="blue bold", no_wrap=True)
    table.add_column("Prompt Line", style="dark_green bold")

    for prompt_file in prompts_files:
        # console.print(f"{prompt_file}")
        try:
            title = os.path.basename(prompt_file)
            with open(prompt_file, 'r') as file:
                lines = file.readlines()
                for lno, line in enumerate(lines):
                    table.add_row(title, f"{lno:03}", line.strip())
                    title = ''

        except Exception as e:
            console.print(f"[bold red]Error parsing file {prompt_file} : {str(e)}[/bold red]")
            console.print_exception()
            sys.exit(1)
        table.add_row('───────────────', '───', '──────────────────────────────────────────────────────────────────────')
    console.print(table)

def get_cmd_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Prompt Engineering Tool.")
    parser.add_argument('-v', '--version', action='store_true', help='Show version information and exit')
    parser.add_argument('--param', nargs=2, action='append',metavar=('key', 'value'),help='Add key/value pairs')
    parser.add_argument('-m', '--models', action='store_true', help='List company models information and exit')
    parser.add_argument('-f', '--functions', action='store_true', help='List functions available to AI and exit')
    parser.add_argument('-p', '--prompts', nargs='?', const='*', help='List Prompts')
    parser.add_argument('-c', '--code', nargs='?', const='*', help='List code in Prompts')
    parser.add_argument('-l', '--list', nargs='?', const='*', help='List Prompt file')
    parser.add_argument('-e', '--execute', nargs='?', const='*', help='Execute one or more Prompts')
    parser.add_argument('-k', '--key', action='store_true', help='Ask for (new) Company Key')
    parser.add_argument('-d', '--debug', nargs='+', choices=['Statements', 'Prompt', 'LLM', 'Functions', 'Messages'],
                        help='Select one or more types of debugging: Statements, Prompt, LLM, Functions, Messages.', default=[])
    parser.add_argument('-r', '--remove', action='store_true', help='remove all .~nn~. files from sub directories')

    return parser.parse_args()

from pathlib import Path


def glob_prompt(prompt_name: str) -> list[Path]:
    if '*' in prompt_name:
        prompt_pattern = Path('prompts') / f"{prompt_name}.prompt"
    else:
        prompt_pattern = Path('prompts') / f"{prompt_name}*.prompt"
    return sorted(Path('.').glob(str(prompt_pattern)))

def main():
    # Ensure 'prompts' directory exists
    if not os.path.exists('prompts'):
        os.makedirs('prompts')

    if not os.path.exists('logs'):
        os.makedirs('logs')

    args = get_cmd_args()
    debug = args.debug
    

    if args.version:
        # Print the version and exit
        console.print(f"[bold cyan]keprompt[/] [bold green]version[/] [bold magenta]{__version__}[/]")
        return

    variables = {}
    if args.param:
        params = {k: v for k, v in args.param}
        # console.print(params)  # print to verify
        variables = params

    # Add in main() after args parsing:
    if args.remove:
        pattern = r'.*\.~\d{2}~\.[^.]+$'
        for root, _, files in os.walk('.'):
            for file in files:
                if re.match(pattern, file):
                    file_path = os.path.join(root, file)
                    try:
                        os.remove(file_path)
                        if debug:
                            log.info(f"Removed {file_path}")
                    except OSError as e:
                        log.error(f"Error removing {file_path}: {e}")
        return

    if args.models:
        # Print the models table and exit
        print_models()
        return

    if args.functions:
        # Print list of functions and exit
        print_functions()
        return

    if args.key:
        get_new_api_key()

    if args.prompts:
        glob_files = glob_prompt(args.prompts)
        if debug: log.info(f"--prompts '{args.prompts}' returned {len(glob_files)} files: {glob_files}")

        if glob_files:
            print_prompt_names(glob_files)
        else:
            log.error(f"[bold red]No Prompt files found for ({args.list})[/bold red]", extra={"markup": True})
        return

    if args.list:
        glob_files = glob_prompt(args.list)
        if debug: log.info(f"--list '{args.list}' returned {len(glob_files)} files: {glob_files}")

        if glob_files:
            print_prompt_lines(glob_files)
        else:
            
            log.error(f"[bold red]No Prompt files found ({args.list})[/bold red]", extra={"markup": True})
        return

    if args.code:
        glob_files = glob_prompt(args.code)
        if debug: log.info(f"--code '{args.code}' returned {len(glob_files)} files: {glob_files}")

        if glob_files:
            print_prompt_code(glob_files)
        else:
            log.error(f"[bold red]No Prompt files found ({args.prompts})[/bold red]", extra={"markup": True})
        return

    if args.execute:
        glob_files = glob_prompt(args.execute)
        if debug: log.info(f"--execute '{args.list}' returned {len(glob_files)} files: {glob_files}")

        if glob_files:
            for prompt_file in glob_files:
                # console.print(params)  # print to verify
                step = VM(prompt_file, args.debug, vdict=variables)
                step.parse_prompt()
                step.execute()
        else:
            log.error(f"[bold red]No execute files found for ({args.execute})[/bold red]", extra={"markup": True})
        return


if __name__ == "__main__":
    main()



import glob
import json
import logging
import os
import sys
import time
from typing import cast, List

import keyring
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from .AiRegistry import AiRegistry, AiModel
from  .keprompt_functions import DefinedFunctions, readfile
from .AiPrompt import AiTextPart, AiImagePart, AiCall, AiResult, AiPrompt, AiMessage, MAX_LINE_LENGTH
from  .keprompt_util import TOP_LEFT, BOTTOM_LEFT, VERTICAL, HORIZONTAL, TOP_RIGHT, RIGHT_TRIANGLE, \
    LEFT_TRIANGLE, \
    HORIZONTAL_LINE, BOTTOM_RIGHT, CIRCLE, backup_file

console = Console()
terminal_width = console.size.width

FORMAT = "%(message)s"
logging.basicConfig(level="NOTSET", format=FORMAT, datefmt="[%X]",
                    handlers=[RichHandler(console=console, rich_tracebacks=True, )])

log = logging.getLogger(__file__)

# Global routines
def print_prompt_code(prompt_files: list[str]) -> None:
    table = Table(title="Execution Messages")
    table.add_column("Prompt", style="cyan bold", no_wrap=True)
    table.add_column("Lno", style="blue bold", no_wrap=True)
    table.add_column("Cmd", style="green bold", no_wrap=True)
    table.add_column("Params", style="dark_green bold")

    for prompt_file in prompt_files:
        # console.print(f"{prompt_file}")
        try:
            vm: VM = VM(prompt_file)
            vm.parse_prompt()
        except Exception as e:
            console.print(f"[bold red]Error parsing file {prompt_file} : {str(e)}[/bold red]")
            console.print_exception()
            sys.exit(1)
        title = os.path.basename(prompt_file)
        if vm.statements:
            for stmt in vm.statements:
                table.add_row(title, f"{stmt.msg_no:03}", stmt.keyword, stmt.value)
                title = ''
            table.add_row('───────────────', '───', '─────────', '──────────────────────────────')
    console.print(table)


class StmtSyntaxError(Exception):
    pass

BEGIN_SUB = '<<'
END_SUB = '>>'

class VM:
    """Class to hold Prompt Virtual Machine execution state"""

    def __init__(self, filename: str, debug: List[str], vdict: dict[str, any] = None):
        self.filename = filename
        self.debug = debug
        self.ip: int = 0
        if vdict:
            self.vdict = vdict
        else:
            self.vdict = dict()
        self.llm: dict[str, any] = dict()
        self.statements: list[StmtPrompt] = []
        # self.messages: list[KeMessage] = []
        self.prompt: AiPrompt = AiPrompt(self)
        self.header: dict[str, any] = {}
        self.data: str = ''
        self.console = Console(width=terminal_width)  # Console for terminal
        self.file_console = None  # Console for file, initialized in execute
        self.model: AiModel = None
        self.model_name: str = ""
        self.company: str = ""
        self.system_value: str = ""
        self.toks_in = 0
        self.cost_in = 0
        self.toks_out = 0
        self.cost_out = 0
        self.total = 0
        self.api_key: str = ''
        self.interaction_no: int = 0

        if debug:
            log.info(f'Instantiated VM(filename="{filename}",debug="{debug}")')

    def print(self, *args, **kwargs):
        """Print method to output to both console and file."""
        self.console.print(*args, **kwargs)  # Print to terminal
        if self.file_console:  # Ensure file is open
            self.file_console.print(*args, **kwargs)  # Print to file

    def debug_print(self, elements: list[str]) -> None:
        """Pretty prints the Virtual Machine class state for debugging"""

        if 'all' in elements:
            elements = ['header', 'llm', 'messages', 'statements', 'variables']

        if 'header' in elements:
            table = Table(title=f"Header Debug Info for {self.filename}")
            table.add_column("VM Property", style="cyan", no_wrap=True, width=35)
            table.add_column("Value", style="green", no_wrap=True)

            table.add_row("Filename", self.filename)
            table.add_row("Debug Options:", str(self.debug))
            table.add_row("IP", str(self.ip))
            table.add_row("url", str(self.llm['url']))
            table.add_row("header", str(self.header))
            table.add_row("data", str(self.data))

            console.print(table)

        # print varname: value
        if 'llm' in elements:
            table = Table(title=f"LLM Debug Info for {self.filename}")

            # Basic info section
            table.add_column("LLM Property", style="cyan", no_wrap=True, width=35)
            table.add_column("Value", style="green", no_wrap=True)

            if self.llm:
                for key, value in self.llm.items():
                    if key == 'API_KEY':
                        value = '... top secret ...'
                    table.add_row(key, str(value))
            else:
                table.add_row("LLM Config", "Not Set")
            console.print(table)

        # Variables dictionary
        # Messages
        if 'messages' in elements:
            table = Table(title=f"Messages Debug Info for {self.filename}")
            # Basic info section
            table.add_column("Mno", style="cyan", no_wrap=True)
            table.add_column("Role", style="blue", no_wrap=True)
            table.add_column("Pno", style="green", no_wrap=True)
            table.add_column("Part", style="green", no_wrap=True, max_width=terminal_width - 25)
            colors = {'user': "[bold steel_blue3]",
                      'assistant': "[bold yellow]",
                      'model': "[bold yellow]",
                      'system': "[bold magenta]",
                      "function": "[bold dark_green]",
                      "result": "[bold dark_green]"}
            if self.prompt:
                for msg_no, msg in enumerate(self.prompt.messages):
                    role = f"{colors[msg.role]}{msg.role}[/]"
                    msg_no_str = f"{msg_no:02}"
                    for pno, part in enumerate(msg.content):
                        part_no = f"{colors[msg.role]}{pno:02}[/]"
                        for substring in str(part).split('\n'):
                            t = f"{colors[msg.role]}{substring}[/]"
                            table.add_row(msg_no_str, role, part_no, t)
                            msg_no_str = ""
                            role = ''
                            part_no = ''
            else:
                table.add_row("", "", "", "Empty")
            console.print(table)

        # Statements
        if 'statements' in elements:
            table = Table(title=f"Statements Debug Info for {self.filename}")
            # Basic info section
            table.add_column("Sno", style="cyan", no_wrap=True)
            table.add_column("Keyword", style="blue", no_wrap=True)
            table.add_column("Value", style="green", no_wrap=True)
            if self.statements:
                last_idx = None
                for idx, stmt in enumerate(self.statements):
                    # input_string = stmt.value.replace('\n', '\\n')
                    hdr = stmt.keyword
                    for substring in stmt.value.split('\n'):
                        if last_idx != idx:
                            str_idx = f"{idx:02}"
                        else:
                            str_idx = ''
                        table.add_row(str_idx, hdr, substring)
                        hdr = ''
                        last_idx = idx
            else:
                table.add_row("00", "", "Empty")
            console.print(table)

        if 'variables' in elements:
            table = Table(title=f"Variables for {self.filename}")
            # Basic info section
            table.add_column("Name", style="cyan", no_wrap=True, width=35)
            table.add_column("Value", style="green", no_wrap=True)
            if self.vdict:
                for key, value in self.vdict.items():
                    table.add_row(key, str(value))
            else:
                table.add_row("Variables", "Empty")
            console.print(table)

    def substitute(self, text: str):

        while END_SUB in text:
            front, back = text.split(END_SUB, 1)
            if BEGIN_SUB not in front:
                return text  # No matching begin marker found

            last_begin = front.rfind(BEGIN_SUB)
            if last_begin == -1:
                return text  # No begin marker found

            # Extract variable name
            variable_name = front[last_begin + 2:]

            # Handle nested dictionaries
            keys = variable_name.split('.')
            value = self.vdict
            try:
                for key in keys:
                    value = value[key]
            except (KeyError, TypeError):
                raise ValueError(f"Variable '{keys}' is not defined")

            # Replace the matched part with the value
            text = front[:last_begin] + str(value) + back

        return text

    def parse_prompt(self) -> None:
        """Parse the prompt file and create a list of statements.
            parse according to rules in docs/PromptLanguage.md
        """

        lines: list[str]

        # read .prompt file
        with open(self.filename, 'r') as file:
            lines = file.readlines()

        # Delete all trailing blank lines
        while lines[-1][0].strip() == '': lines.pop()

        for lno, line in enumerate(lines):
            try:
                line = line.strip()  # remove trailing blanks
                if not line: continue  # skip blank lines

                # Get Keyword and Value in all cases.

                if line[0] != '.':  # No Dot in col 1
                    keyword, value = '.text', line
                else:
                    # has '.' in col 1
                    if ' ' in line:  # has space therefore has .keyword<space>value
                        keyword, value = line.split(' ', 1)
                    else:  # No space therefore only .keyword
                        keyword, value = line, ''

                    if keyword not in keywords:  # last case have .keyword but it is not a valid keyword
                        keyword, value = '.text', line

                # okay concatenate .text
                if lno and keyword == '.text':
                    last = self.statements[-1]
                    if last.keyword in ['.assistant', '.system', '.text', '.user']:
                        last.value = f"{last.value}\n{value}".strip()
                        continue

                self.statements.append(make_statement(self, len(self.statements), keyword=keyword, value=value))

            except Exception as e:
                raise StmtSyntaxError(
                    f"{VERTICAL} [red]Error parsing file {self.filename}:{lno} error: {str(e)}.[/]\n\n")

        # Implicit .exec
        if lines[-1][0:5] != '.exec':
            self.statements.append(make_statement(self, len(self.statements), keyword='.exec', value=''))

        return

    def print_exception(self) -> None:
        """Print exception information to both console and file outputs."""
        self.console.print()
        self.console.print_exception(show_locals=True, width=terminal_width)  # Print to terminal
        if self.file_console:  # Ensure file is open
            self.file_console.print_exception()  # Print to file

    def load_llm(self, parms: dict[str, str]) -> None:

        if 'model' not in parms:
            raise StmtSyntaxError(f".llm syntax error: model not defined")
        self.model_name = parms['model']

        if self.model_name not in AiRegistry.models:
            raise StmtSyntaxError(f"Not Defined Error: Model {self.model_name} is not defined")
        self.model = AiRegistry.get_model(self.model_name)

        if self.model.company == '':
            raise StmtSyntaxError(f"Bad Model Definition error: company not defined for model {self.model_name}")
        self.company = self.model.company

        # copy parms to vdict
        for k, v in parms.items():
            self.vdict[k] = v

        self.vdict['company'] = self.company
        self.vdict['filename'] = self.filename
        self.vdict['model'] = self.model

    def execute(self) -> None:
        """Execute the statements in the prompt file"""
        if self.debug and 'Prompt' in self.debug: log.info(f'execute({self.filename} with {len(self.statements)} statements)')

        base_name = os.path.splitext(os.path.basename(self.filename))[0]
        logfile_name = backup_file(f"logs/{base_name}.log", backup_dir='logs', extension='.log')
        with open(logfile_name, 'w') as file:
            self.file_console = Console(file=file, record=True)  # Open file for writing

            self.print(
                f"[bold white]{TOP_LEFT}{HORIZONTAL * 2}[/][bold white]{os.path.basename(self.filename):{HORIZONTAL}<{terminal_width - 4}}{TOP_RIGHT}[/]"
            )

            for stmt_no, stmt in enumerate(self.statements):
                try:
                    stmt.execute(self)
                except Exception as e:
                    self.print(f"[bold red]Error executing statement above : {str(e)}\n\n")
                    self.print_exception()
                    sys.exit(9)

                if stmt.keyword == '.exit':
                    break

            self.print(f"{BOTTOM_LEFT}{HORIZONTAL * (terminal_width - 2)}{BOTTOM_RIGHT}")

            self.file_console.file.close()  # Close file console at end
            logfile_name_html = backup_file(f"logs/{base_name}.svg", backup_dir='logs', extension='.svg')
            self.console.save_svg(logfile_name_html)
            console.print(f"Wrote {logfile_name_html} to disk")

    def print_with_wrap(self, is_response: bool, line: str) -> None:
        line_len = terminal_width - 23

        color = '[bold green]'
        if is_response:
            color = '[bold blue]'

        print_line = line.replace('\n', '\\n')[:line_len]  # Truncate if longer
        print_line = f"{print_line:<{line_len + 8}}"  # Ensure it is exactly line_len wide with spaces

        if is_response:
            hdr = f"[bold white]{VERTICAL}[/]{color}   {LEFT_TRIANGLE}{HORIZONTAL_LINE * 5}{CIRCLE}  "
        else:
            hdr = f"[bold white]{VERTICAL}[/]{color}   {CIRCLE}{HORIZONTAL_LINE * 5}{RIGHT_TRIANGLE}  "

        self.print(f"{hdr}[/]:{print_line}[bold white]{VERTICAL}[/]")

    def log_conversation(self):
        base_name = os.path.splitext(os.path.basename(self.filename))[0]
        logfile_name = backup_file(f"logs/{base_name}_messages.json", backup_dir='logs', extension='.json')
        with open(logfile_name, 'w') as file:
            file.write(json.dumps(self.prompt.to_json()))

    def log_last_json(self, data: dict[str, any]):
        base_name = os.path.splitext(os.path.basename(self.filename))[0]
        logfile_name = backup_file(f"logs/{base_name}_last_msgs.json", backup_dir='logs', extension='.json')
        with open(logfile_name, 'w') as file:
            file.write(json.dumps(data, indent=2, sort_keys=True))

    def print_json(self,    label: str, data: dict) -> None:
        """Print a JSON object to the console"""
        self.print(f"{label}:")
        pdict = {}
        for k, v in data.items():
            if isinstance(v, str):
                pdict[k] = v.replace('\n', '\\n')
                if len(v) > MAX_LINE_LENGTH:
                    pdict[k] = f"{v[:MAX_LINE_LENGTH - 3]}..."
                else:
                    pdict[k] = v
            else:
                pdict[k] = v
        self.print(json.dumps(pdict, indent=2, sort_keys=True))

class StmtPrompt:

    def __init__(self, vm: VM, msg_no: int, keyword: str, value: str):
        self.msg_no = msg_no
        self.keyword = keyword
        self.value = value
        self.vm = vm

    def console_str(self) -> str:
        line_len = terminal_width - 14
        header = f"[bold white]{VERTICAL}[/][white]{self.msg_no:02}[/] [cyan]{self.keyword:<8}[/] "
        value = self.value
        if len(value) == 0:
            value = " "
        lines = value.split("\n")

        rtn = ""
        for line in lines:
            while len(line) > 0:
                print_line = f"{line:<{line_len}}[bold white]{VERTICAL}[/]"
                rtn = f"{rtn}\n{header}[green]{print_line}[/]"
                header = f"[bold white]{VERTICAL}[/]            "
                line = line[line_len:]

        return rtn[1:]

    def __str__(self):
        return self.console_str()

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())


class StmtAssistant(StmtPrompt):

    def execute(self, vm: VM) -> None:
        if vm.debug and 'Statements' in vm.debug:
            vls = self.value.split('\n')
            vl = vls.pop(0)
            vm.print(f"[bold white]{VERTICAL}[/][white]{self.msg_no:02}[/] [cyan]{self.keyword:<8}[/] [green]{vl}[/]")
            for vl in vls:
                vm.print(f"[bold white]{VERTICAL}[/]            [green]{vl}[/]")
        vm.print(self.console_str())
        if not self.value:
            vm.prompt.add_message(vm=vm, role='assistant', content=[])
        else:
            vm.prompt.add_message(vm=vm, role='assistant', content=[AiTextPart(vm=vm, text=self.value)])


class StmtClear(StmtPrompt):

    def execute(self, vm: VM) -> None:
        if vm.debug and 'Statements' in vm.debug:
            vm.print(
                f"[bold white]{VERTICAL}[/][white]{self.msg_no:02}[/] [cyan]{self.keyword:<8}[/] [green]{self.value}[/]")

        try:
            parms = json.loads(self.value)
        except Exception as e:
            vm.print(f"{VERTICAL} [white on red]Error parsing .clear parameters: {str(e)}[/]\n\n")
            vm.print_exception()
            sys.exit(9)
            # raise PromptSyntaxError(f"Error parsing .clear parameters: {str(e)}")

        if not isinstance(parms, list):
            vm.print(
                f"{VERTICAL} [white on red]Error parsing .clear parameters expected list, but got {type(parms).__name__}: {self.value}")
            sys.exit(9)

        for k in parms:
            try:
                log_files = glob.glob(k)  # Use glob to find all files matching the pattern

                for file_path in log_files:
                    if os.path.isfile(file_path):  # Ensure that it's a file
                        if vm.debug and 'Statements' in vm.debug:
                            vm.print(f"{VERTICAL} [bold green] Deleting {k}[/bold green]")
                        try:
                            os.remove(file_path)
                            vm.print(f"File {file_path} deleted successfully.")
                        except OSError as e:
                            vm.print(f"Error deleting file {file_path}: {str(e)}")

                    if vm.debug and 'Statements' in vm.debug:
                        vm.print(f"{VERTICAL} [bold green]File {k} deleted successfully.[/bold green]")
            except OSError as e:
                vm.print(f"{VERTICAL} [white or red]Error deleting file {k}: {str(e)}[/]\n\n")


class StmtCmd(StmtPrompt):

    def execute(self, vm: VM) -> None:
        """Execute a command that was defined in a prompt file (.prompt)"""

        function_name, args = self.value.split('(', maxsplit=1)
        args = args[:-1]
        args_list = args.split(",")
        function_args = {}

        if function_name == 'askuser':
            vm.print(self.console_str() + ': ', end='')
        else:
            vm.print(self.console_str())

        for arg in args_list:
            name, value = arg.split("=", maxsplit=1)
            function_args[name] = value

        if function_name not in DefinedFunctions:
            vm.print(
                f"[bold red]Error executing {function_name}({function_args}): {function_name} is not defined.[/bold red]")
            raise Exception(f"{function_name} is not defined.")

        try:
            text = DefinedFunctions[function_name](**function_args)
        except Exception as err:
            vm.print(f"Error executing {function_name}({function_args})): {str(err)}")
            raise err

        last_msg = vm.prompt.messages[-1]
        last_msg.content.append(AiTextPart(vm=vm, text=text))


class StmtComment(StmtPrompt):
    pass


class StmtDebug(StmtPrompt):

    def execute(self, vm: VM) -> None:

        vm.print(self.console_str())

        if not self.value:
            self.value = '["all"]'

        if self.value[0] != '[':
            self.value = f"[{self.value}]"

        # vm.print(self.value)
        try:
            parms = json.loads(self.value)
        except Exception as e:
            vm.print(f"{VERTICAL} [white on red]Error parsing .debug parameters: {str(e)}[/]\n\n")
            vm.print_exception()
            sys.exit(9)

        if not isinstance(parms, list):
            vm.print(
                f"{VERTICAL} [white on red]Error parsing .debug parameters expected list, but got {type(parms).__name__}: {self.value}")
            sys.exit(9)

        vm.debug_print(elements=parms)


class StmtExec(StmtPrompt):

    def execute(self, vm: VM) -> None:
        """Execute a request to an LLM"""
        header = f"[bold white]{VERTICAL}[/][white]{self.msg_no:02}[/] [cyan]{self.keyword:<8}[/]"
        # vm.print(header, end='')

        start_time = time.time()
        response: AiMessage = vm.prompt.ask(label=header)
        elapsed_time = time.time() - start_time

        pline = f"{header} {elapsed_time:.2f} secs output tokens {vm.prompt.toks_out} at {vm.prompt.toks_out / elapsed_time:.2f} tps"
        used_bytes = 13 + 11 + len(vm.company) + 2 + len(vm.model_name) + 9
        no_bytes_remaining = terminal_width - used_bytes
        vm.print(f"{pline:<{no_bytes_remaining}}[bold white]{VERTICAL}[/]")

        if vm.debug and 'Statements' in vm.debug:
            vm.print(f"[bold blue]Response from {vm.company} API:[/bold blue]")
            vm.print(response)

        pline = f"Tokens In={vm.toks_in}(${vm.cost_in:06.4f}), Out={vm.toks_out}(${vm.cost_out:06.4f}) Total=${vm.total:06.4f}"
        vm.print(f"{header}{pline:<{terminal_width - 14}}[bold white]{VERTICAL}[/]")

        vm.log_conversation()


class StmtExit(StmtPrompt):

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())


class StmtInclude(StmtPrompt):
    # Read a file and add its content to last_msg

    def execute(self, vm: VM) -> None:
        filename = vm.substitute(self.value)
        vm.print(self.console_str())
        lines = readfile(filename=filename)
        last_msg = vm.prompt.messages[-1]
        last_msg.content.append(AiTextPart(vm=vm, text=lines))


class StmtImage(StmtPrompt):
    # Read an Image file and add its content to last_msg

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())
        filename = self.value
        vm.prompt.add_message(vm=vm, role="user", content=[AiImagePart(vm=self.vm, filename=filename)])


class StmtLlm(StmtPrompt):

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())
        try:
            if vm.llm:
                raise (StmtSyntaxError(f".llm syntax: only one .lls statement allowed in vm {vm.filename}"))

            if self.value[0] != '{':
                self.value = "{" + self.value + "}"

            value = self.vm.substitute(self.value)

            try:
                parms = json.loads(value)
            except Exception as e:
                vm.print(f"{VERTICAL} [white on red]Error parsing .llm parameters: {str(e)}[/]\n\n")
                vm.print_exception()
                sys.exit(9)

            if not isinstance(parms, dict):
                raise (StmtSyntaxError(
                    f".llm syntax: parameters expected dict, but got {type(parms).__name__}: {self.value}"))

            if 'model' not in parms:
                raise (StmtSyntaxError(f".llm syntax:  'model' parameter is required but missing {self.value}"))

            vm.load_llm(parms)

        except Exception as err:
            vm.print_exception()
            sys.exit(9)

        # Now we that we have loaded the LLM,  we will load the API_KEY
        try:
            api_key = keyring.get_password('keprompt', username=vm.company)
        except keyring.errors.PasswordDeleteError:
            vm.print(f"[bold red]Error accessing keyring ('keprompt', username={vm.company})[/bold red]")
            api_key = None

        if api_key is None:
            api_key = console.input(f"Please enter your {vm.company} API key: ")
            keyring.set_password("keprompt", username=vm.company, password=api_key)
        if not api_key:
            vm.print("[bold red]API key cannot be empty.[/bold red]")
            sys.exit(1)

        vm.llm['API_KEY'] = api_key
        vm.api_key = api_key
        vm.prompt.api_key = vm.api_key
        vm.prompt.company = vm.company
        vm.prompt.model = vm.model_name


class StmtSystem(StmtPrompt):

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())
        if not self.value:
            vm.prompt.add_message(vm=vm, role='system', content=[])
        else:
            vm.prompt.add_message(vm=vm, role='system', content=[AiTextPart(vm=vm, text=self.value)])


class StmtText(StmtPrompt):

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())
        if vm.prompt.messages[-1].role in ['assistant', 'system', 'user']:
            vm.prompt.messages[-1].content.append(AiTextPart(vm=vm, text=self.value))
        else:
            vm.prompt.add_message(vm=vm, role='user', content=[AiTextPart(vm=vm, text=self.value)])


class StmtUser(StmtPrompt):

    def execute(self, vm: VM) -> None:
        vm.print(self.console_str())
        if not self.value:
            vm.prompt.add_message(vm=vm, role='user', content=[])
        else:
            vm.prompt.add_message(vm=vm, role='user', content=[AiTextPart(vm=vm, text=self.value)])


# Create a _PromptStatement subclass depending on keyword
StatementTypes: dict[str, type(StmtPrompt)] = {
    '.#': StmtComment,
    '.assistant': StmtAssistant,
    '.clear': StmtClear,
    '.cmd': StmtCmd,
    '.debug': StmtDebug,
    '.exec': StmtExec,
    '.exit': StmtExit,
    '.image': StmtImage,
    '.include': StmtInclude,
    '.llm': StmtLlm,
    '.system': StmtSystem,
    '.text': StmtText,
    '.user': StmtUser,
}

keywords = StatementTypes.keys()


def make_statement(vm: VM, msg_no: int, keyword: str, value: str) -> StmtPrompt:
    my_class = StatementTypes[keyword]
    return my_class(vm, msg_no, keyword, value)



import platform
import subprocess
import sys
from typing import Any, Dict

from rich.console import Console
from rich.prompt import Prompt
from rich.theme import Theme

from .keprompt_util import backup_file

console = Console()

# Define custom theme for prompts
theme = Theme({"prompt": "bold blue", "answer": "italic cyan"})
question_console = Console(theme=theme)


def get_webpage_content(url: str) -> str:
    """
    Fetches the content of a webpage and converts it to text.

    Args:
        url (str): The URL of the webpage to fetch.

    Returns:
        str: The text content of the webpage.

    Raises:
        Exception: If there is an error fetching the URL.
    """
    command = f"wget2 --content-on-error -O - {url} | html2text"

    try:
        process = subprocess.run(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            text=True,
        )
        return process.stdout
    except subprocess.CalledProcessError as e:
        error_msg = e.stderr.strip() or e.stdout.strip()
        raise Exception(f"Error fetching URL '{url}': {error_msg}") from e


def readfile(filename: str) -> str:
    """
    Reads the contents of a local file.

    Args:
        filename (str): The name of the file to read.

    Returns:
        str: The contents of the file.

    Exits:
        Exits the program if the file cannot be read.
    """
    try:
        with open(filename, 'r', encoding='utf-8') as file:
            return file.read()
    except Exception as err:
        console.print(f"Error accessing file '{filename}': {err}\n\n", style="bold red")
        console.print_exception()
        sys.exit(9)


def askuser(question: str) -> str:
    """
    Prompts the user for input with enhanced formatting.

    Args:
        question (str): The question to present to the user.

    Returns:
        str: The user's response.
    """
    return Prompt.ask(f"[prompt]{question}[/prompt]", console=question_console)


def wwwget(url: str) -> Any:
    """
    Retrieves the content of a webpage.

    Args:
        url (str): The URL of the webpage to retrieve.

    Returns:
        str: The content of the webpage or an error dictionary.
    """
    try:
        return get_webpage_content(url)
    except Exception as err:
        console.print(f"Error while retrieving URL '{url}': {err}", style="bold red")
        return {
            'role': "function",
            'name': 'wwwget',
            'content': f'ERROR: URL not returned: {url}'
        }


def writefile(filename: str, content: str) -> str:
    """
    Writes content to a file with versioning.

    Args:
        filename (str): The name of the file to write.
        content (str): The content to write to the file.

    Returns:
        str: The path of the written file.
    """
    new_filename = backup_file(filename)

    try:
        with open(new_filename, 'w', encoding='utf-8') as f:
            f.write(content)
        return f"Content written to file '{new_filename}'"
    except Exception as e:
        console.print(f"Failed to write to file '{new_filename}': {e}", style="bold red")
        raise

import base64

def write_base64_file(filename: str, base64_str: str) -> str:
    """
    Decodes a base64 string and writes the decoded content to a file with versioning.

    Args:
        filename (str): The name of the file to write.
        base64_str (str): The base64 encoded content to write to the file.

    Returns:
        str: The path of the written file.
    """
    new_filename = backup_file(filename)

    try:
        decoded_content = base64.b64decode(base64_str)
        with open(new_filename, 'wb') as f:
            f.write(decoded_content)
        return f"Content written to file '{new_filename}'"
    except Exception as e:
        console.print(f"Failed to write to file '{new_filename}': {e}", style="bold red")
        raise

def execcmd(cmd: str) -> str:
    """
    Executes a shell command and returns its output.

    Args:
        cmd (str): The command to execute.

    Returns:
        str: The standard output or error message.
    """
    sanitized_cmd = cmd.strip('\"\'') if cmd and cmd[0] in {'"', "'"} else cmd

    try:
        result = subprocess.run(
            ['/bin/sh', '-c', sanitized_cmd],
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        return f"Error executing command: {e.stderr.strip()}"
    except Exception as e:
        return f"Unexpected error: {str(e)}"


# Determine the operating system descriptor
os_descriptor = platform.platform()

# Define available tools
DefinedToolsArray = [
    {   'type': 'function',
        'function': {
            "name": "readfile",
            "description": "Read the contents of a named file",
            "parameters": {
                "type": "object",
                "properties": {
                    "filename": {
                        "type": "string",
                        "description": "The name of the file to read",
                    },
                },
                "required": ["filename"],
                "additionalProperties": False
            },
        }
    },
    {
        'type': 'function',
        'function': {
            "name": "wwwget",
            "description": "Read a webpage URL and return the contents",
            "parameters": {
                "type": "object",
                "properties": {
                    "url": {
                        "type": "string",
                        "description": "The URL of the web page to read",
                    },
                },
                "required": ["url"],
                "additionalProperties": False
            },
        }
    },
    {   'type': 'function',
        'function': {
            "name": "writefile",
            "description": "Write the contents to a named file on the local file system",
            "parameters": {
                "type": "object",
                "properties": {
                    "filename": {
                        "type": "string",
                        "description": "The name of the file to write",
                    },
                    "content": {
                        "type": "string",
                        "description": "The content to be written to the file",
                    },
                },
                "required": ["filename", "content"],
                "additionalProperties": False
            },
        }
    },
    {
        'type': 'function',
        'function': {
            "name": "execcmd",
            "description": f"Execute a command on the local {os_descriptor} system",
            "parameters": {
                "type": "object",
                "properties": {
                    "cmd": {
                        "type": "string",
                        "description": "Command to be executed",
                    },
                },
                "required": ["cmd"],
                "additionalProperties": False
            },
        }
    },
    {   'type': 'function',
        'function': {
            "name": "askuser",
            "description": "Get clarification by asking the user a question",
            "parameters": {
                "type": "object",
                "properties": {
                    "question": {
                        "type": "string",
                        "description": "Question to ask the user",
                    },
                },
                "required": ["question"],
                "additionalProperties": False
            },
        }
    },

    {   'type': 'function',
         'function': {
             "name": "write_base64_file",
             "description": "Decode base64 content and write the decoded data to a named file on the local file system",
             "parameters": {
                 "type": "object",
                 "properties": {
                     "filename": {
                         "type": "string",
                         "description": "The name of the file to write",
                     },
                    "base64_str": {
                        "type": "string",
                        "description": "The base64 encoded content to be decoded and written to the file",
                    },
                 },
                 "required": ["filename", "base64_str"],
                 "additionalProperties": False
                },
        }
    }
]

# Mapping of function names to their implementations
DefinedFunctions: Dict[str, Any] = {
    "readfile": readfile,
    "wwwget": wwwget,
    "writefile": writefile,
    "execcmd": execcmd,
    "askuser": askuser,
    "write_base64_file": write_base64_file
}



# CompanyAi.py
import abc
from typing import List, Dict, Any, TYPE_CHECKING

import requests
from rich import json
from rich.console import Console
from rich.progress import TimeElapsedColumn, Progress

from .keprompt_functions import DefinedFunctions
from .keprompt_util import VERTICAL

console = Console()
terminal_width = console.size.width

if TYPE_CHECKING:
    from .AiPrompt import AiMessage, AiPrompt, AiCall, AiResult


class AiCompany(abc.ABC):

    def __init__(self, prompt: 'AiPrompt'):
        self.prompt = prompt
        self.system_prompt = None



    @abc.abstractmethod
    def prepare_request(self, messages: List[Dict]) -> Dict:
        """Override to create company-specific request format"""
        pass

    @abc.abstractmethod
    def get_api_url(self) -> str:
        """Override to provide company API endpoint"""
        pass

    @abc.abstractmethod
    def get_headers(self) -> Dict:
        """Override to provide company-specific headers"""
        pass

    @abc.abstractmethod
    def to_company_messages(self, messages: List['AiMessage']) -> List[Dict]:
        pass

    @abc.abstractmethod
    def to_ai_message(self, response: Dict) -> 'AiMessage':
        """Convert full API response to AiMessage. Each company implements their response parsing."""
        pass

    def call_llm(self, label: str) -> List['AiMessage']:
        do_again = True
        responses = []
        call_count = 0

        self.prompt.vm.print(f"{label} [bold blue]Calling {self.prompt.company}::{self.prompt.model}[/bold blue]")

        while do_again:
            call_count += 1
            do_again = False

            if 'Messages' in self.prompt.vm.debug:
                self.prompt.print_messages(f"Sent to {self.prompt.model}")

            company_messages = self.to_company_messages(self.prompt.messages)
            request = self.prepare_request(company_messages)

            # Make API call with formatted label
            call_label = f"Call-{call_count:02d}"
            response = self.make_api_request(
                url=self.get_api_url(),
                headers=self.get_headers(),
                data=request,
                label=call_label
            )

            response_msg = self.to_ai_message(response)
            self.prompt.messages.append(response_msg)
            responses.append(response_msg)

            tool_msg = self.call_functions(response_msg)
            if tool_msg:
                do_again = True
                self.prompt.messages.append(tool_msg)
                responses.append(tool_msg)

        if 'Messages' in self.prompt.vm.debug:
            self.prompt.print_messages(f"Received from {self.prompt.model}")

        return responses


    def call_functions(self, message):
        # Import here to avoid Circular Imports
        from .AiPrompt import AiResult, AiMessage, AiCall

        tool_results = []

        for part in message.content:
            if not isinstance(part, AiCall): continue

            try:
                if 'Functions' in self.prompt.vm.debug:
                    self.prompt.vm.print(f"[bold green]Executing function: {part.name} with args: {part.arguments}[/]")

                result = DefinedFunctions[part.name](**part.arguments)

                if 'Functions' in self.prompt.vm.debug:
                    self.prompt.vm.print(f"[bold green]Function Return: {result}[/]")

                tool_results.append(AiResult(vm=self.prompt.vm, name=part.name, id=part.id or "", result=str(result)))
            except Exception as e:
                tool_results.append(AiResult(vm=self.prompt.vm, name=part.name, id=part.id or "", result=f"Error calling {str(e)}"))

        return AiMessage(vm=self.prompt.vm, role="tool", content=tool_results) if tool_results else None



    def make_api_request(self, url: str, headers: Dict, data: Dict, label: str) -> Dict:

        if 'LLM' in self.prompt.vm.debug:
            self.prompt.vm.print_json(label=f"[bold blue]Sending to {self.prompt.company}::{self.prompt.model}[/bold blue]", data=data)

        with Progress("[progress.description]{task.description}", TimeElapsedColumn(), console=console, transient=True, ) as progress:
            call_label = f"[white]{VERTICAL}[/]{' ' * 12}{label}"
            task = progress.add_task(description=call_label, total=None)
            response = requests.post(url=url, headers=headers, json=data)
            progress.update(task, description=f"[white]{VERTICAL}[/]{' ' * 12}{label}")

        if response.status_code != 200:
            raise Exception(f"{self.prompt.company}::{self.prompt.model} API error: {response.text}")

        resp_obj = response.json()

        tokens = resp_obj.get("usage", {}).get("output_tokens", 0)
        elapsed = response.elapsed.total_seconds()
        tokens_per_sec = tokens / elapsed if elapsed > 0 else 0
        timings = f"Elapsed: {elapsed:.2f} seconds {tokens_per_sec:.2f} tps"
        head = f"{label} {timings}"
        rem = terminal_width - len(head)
        self.prompt.vm.print(f"[white]{VERTICAL}{' ' * 12}{head}{' ' * rem}[white]{VERTICAL}[/]", end='')
        # console.print(f"{head}{' ' * rem} [white]{VERTICAL}[/]", end='')

        retval = response.json()

        if 'LLM' in self.prompt.vm.debug:
            self.prompt.vm.print_json(label=f"[bold blue]Received from {self.prompt.company}::{self.prompt.model}[/bold blue]", data=retval)

        # Update token counts
        self.prompt.toks_in += retval.get("usage", {}).get("input_tokens", 0)
        self.prompt.toks_out += retval.get("usage", {}).get("output_tokens", 0)

        return retval



from typing import Dict, List
import json
from rich.console import Console

from .AiRegistry import AiRegistry
from .AiCompany import AiCompany
from .AiPrompt import AiMessage, AiTextPart, AiCall
from .keprompt_functions import DefinedToolsArray

console = Console()
terminal_width = console.size.width


class AiOpenAi(AiCompany):
    def prepare_request(self, messages: List[Dict]) -> Dict:
        return {
            "model": self.prompt.model,
            "messages": messages,
            "tools": DefinedToolsArray
        }

    def get_api_url(self) -> str:
        return "https://api.openai.com/v1/chat/completions"

    def get_headers(self) -> Dict:
        return {
            "Authorization": f"Bearer {self.prompt.api_key}",
            "Content-Type": "application/json"
        }

    def to_ai_message(self, response: Dict) -> AiMessage:
        choice = response.get("choices", [{}])[0].get("message", {})
        content = []

        if choice.get("content"):
            content.append(AiTextPart(vm=self.prompt.vm, text=choice["content"]))

        for tool_call in choice.get("tool_calls", []):
            content.append(AiCall(vm=self.prompt.vm,name=tool_call["function"]["name"],arguments=tool_call["function"]["arguments"],id=tool_call["id"]))

        return AiMessage(vm=self.prompt.vm, role="assistant", content=content)

    def to_company_messages(self, messages: List[AiMessage]) -> List[Dict]:
        openai_messages = []

        for msg in messages:
            if msg.role == "system":
                self.system_message = msg.content[0].text if msg.content else None
                continue

            content = []
            tool_calls = []
            tool_results = {}

            for part in msg.content:
                if   part.type == "text":       content.append({"type": "text", "text": part.text})
                elif part.type == "image_url":  content.append({'type': 'image_url','image_url': {'url': f"data:{part.media_type};base64,{part.file_contents}"}})
                elif part.type == "call":       tool_calls.append({'id': part.id,'type': 'function','function': {'name': part.name,'arguments': json.dumps(part.arguments)}})
                elif part.type == 'result':     tool_results= {'role': "tool", 'tool_call_id': part.id,'content': part.result}
                else:                           raise ValueError(f"Unknown part type: {part.type}")

            if msg.role == "tool":
                message = tool_results
            else:
                message = {"role": msg.role,"content": content[0]["text"] if len(content) == 1 else content}
                if tool_calls:
                    message["tool_calls"] = tool_calls

            openai_messages.append(message)

        return openai_messages


# Register handler and models
AiRegistry.register_handler(company_name="OpenAI", handler_class=AiOpenAi)

OpenAI_Models = {
    "gpt-4o-mini": {"company": "OpenAI", "model": "gpt-4o-mini", "input": 0.00000015, "output": 0.0000006, "context": 128000},
    "gpt-4o": {"company": "OpenAI", "model": "gpt-4o", "input": 0.000005, "output": 0.00002, "context": 128000},
    "gpt-4o-2024-05-13": {"company": "OpenAI", "model": "gpt-4o-2024-05-13", "input": 0.000005, "output": 0.000015, "context": 128000},
    "gpt-4o-mini-2024-07-18": {"company": "OpenAI", "model": "gpt-4o-mini-2024-07-18", "input": 0.00000015, "output": 0.0000006, "context": 128000},
    "gpt-4o-2024-08-06": {"company": "OpenAI", "model": "gpt-4o-2024-08-06", "input": 0.000005, "output": 0.00002, "context": 128000},
    "o1": {"company": "OpenAI", "model": "o1", "input": 0.000003, "output": 0.000012, "context": 128000},
    "o1-mini": {"company": "OpenAI", "model": "o1-mini", "input": 0.0000006, "output": 0.0000024, "context": 128000},
    "o3-mini": {"company": "OpenAI", "model": "o3-mini", "input": 0.0000011, "output": 0.0000044, "context": 128000}
}

AiRegistry.register_models_from_dict(model_definitions=OpenAI_Models)


from typing import Dict, List
from rich.console import Console

from .AiRegistry import AiRegistry
from .AiCompany import AiCompany
from .AiPrompt import AiMessage, AiTextPart, AiCall
from .keprompt_functions import DefinedToolsArray

console = Console()
terminal_width = console.size.width


class AiMistral(AiCompany):
    def prepare_request(self, messages: List[Dict]) -> Dict:
        return {"model": self.prompt.model,"messages": messages,"tools": DefinedToolsArray,"tool_choice": "auto"}

    def get_api_url(self) -> str:
        return "https://api.mistral.ai/v1/chat/completions"

    def get_headers(self) -> Dict:
        return {"Authorization": f"Bearer {self.prompt.api_key}","Content-Type": "application/json","Accept": "application/json"}

    def to_ai_message(self, response: Dict) -> AiMessage:
        choice = response.get("choices", [{}])[0].get("message", {})
        content = []

        if choice.get("content"):
            content.append(AiTextPart(vm=self.prompt.vm, text=choice["content"]))

        tool_calls = choice.get("tool_calls", [])
        if not tool_calls:
            tool_calls = []

        for tool_call in tool_calls:
            content.append(AiCall(vm=self.prompt.vm,name=tool_call["function"]["name"],arguments=tool_call["function"]["arguments"],id=tool_call["id"]))

        return AiMessage(vm=self.prompt.vm, role="assistant", content=content)

    def to_company_messages(self, messages: List[AiMessage]) -> List[Dict]:
        mistral_messages = []

        for msg in messages:
            if msg.role == "system":
                self.system_message = msg.content[0].text if msg.content else None
                continue

            content = []
            tool_calls = []
            tool_results = {}

            for part in msg.content:
                if   part.type == "text":       content.append({"type": "text", "text": part.text})
                elif part.type == "image_url":  content.append({'type': 'image_url','image_url': {'url': f"data:{part.media_type};base64,{part.file_contents}"}})
                elif part.type == "call":       tool_calls.append({'id': part.id,'type': 'function','function': {'name': part.name,'arguments': part.arguments}})
                elif part.type == 'result':     tool_results = {'id': part.id,'content': part.result}
                else:                           raise ValueError(f"Unknown part type: {part.type}")


            if msg.role == "tool":
                message = {"role": "tool", "content": tool_results["content"], "tool_call_id": tool_results["id"]}
            else:
                message = {"role": msg.role,"content": content}
                if tool_calls:
                    message["tool_calls"] = tool_calls

            mistral_messages.append(message)

        return mistral_messages


# Register handler and models
AiRegistry.register_handler(company_name="MistralAI", handler_class=AiMistral)

Mistral_Models = {
    "mistral-large-latest": {"company": "MistralAI","model": "mistral-large-latest","input": 0.000002,"output": 0.000006,"context": 32000},
    "mistral-small-latest": {"company": "MistralAI","model": "mistral-small-latest","input": 0.0000002,"output": 0.0000006,"context": 32000},
    "codestral-latest":     {"company": "MistralAI","model": "codestral-latest",    "input": 0.0000003,"output": 0.0000009,"context": 32000},
    "ministral-8b-latest":  {"company": "MistralAI","model": "ministral-8b-latest", "input": 0.0000001,"output": 0.0000001,"context": 32000},
    "ministral-3b-latest":  {"company": "MistralAI","model": "ministral-3b-latest", "input": 0.00000004,"output": 0.00000004,"context": 32000},
    "pixtral-large-latest": {"company": "MistralAI","model": "pixtral-large-latest","input": 0.0000002,"output": 0.0000006,"context": 32000},
    "pixtral-12b":          {"company": "MistralAI","model": "pixtral-12b",         "input": 0.00000015,"output": 0.00000015,"context": 32000
    }
}

AiRegistry.register_models_from_dict(model_definitions=Mistral_Models)




import time
from typing import Dict, List
import json
from rich.console import Console

from .AiRegistry import AiRegistry
from .AiCompany import AiCompany
from .AiPrompt import AiMessage, AiTextPart, AiCall, AiResult, AiPrompt
from .keprompt_functions import DefinedFunctions, DefinedToolsArray


console = Console()
terminal_width = console.size.width


class AiAnthropic(AiCompany):

    def prepare_request(self, messages: List[Dict]) -> Dict:
        return {
            "model": self.prompt.model,
            "messages": messages,
            "tools": AnthropicToolsArray,
            "max_tokens": 4096
        }

    def get_api_url(self) -> str:
        return "https://api.anthropic.com/v1/messages"

    def get_headers(self) -> Dict:
        return {
            "x-api-key": self.prompt.api_key,
            "anthropic-version": "2023-06-01",
            "content-type": "application/json"
        }

    def to_ai_message(self, response: Dict) -> 'AiMessage':
        content = []
        resp_content = response.get("content", [])

        for part in resp_content:
            if part["type"] == "text":
                content.append(AiTextPart(vm=self.prompt.vm, text=part["text"]))
            elif part["type"] == "tool_use":
                content.append(AiCall(vm=self.prompt.vm, id=part["id"],name=part["name"], arguments=part["input"]))

        return AiMessage(vm=self.prompt.vm, role="assistant", content=content)
    def to_company_messages(self, messages: List) -> List[Dict]:

        company_mesages = []
        for msg in messages:
            content = []
            if msg.role == "system":
                self.system_message = msg.content[0].text if msg.content else None
            else:
                for part in msg.content:
                    if   part.type == "text":       content.append({'type': 'text', 'text': part.text})
                    elif part.type == "image_url":  content.append({'type': 'image', 'source': {'type': 'base64', 'media_type': part.media_type, 'data': part.file_contents}})
                    elif part.type == "call":       content.append({'type': 'tool_use', 'id': part.id, 'name': part.name, 'input': part.arguments})
                    elif part.type == 'result':     content.append({'type': 'tool_result', 'tool_use_id': part.id, 'content': part.result})
                    else: raise Exception(f"Unknown part type: {part.type}")

                role = "assistant" if msg.role == "assistant" else "user"
                company_mesages.append({"role": role, "content": content})

        return company_mesages


# Prepare tools for Anthropic and Google integrations
AnthropicToolsArray = [
    {
        "name": tool['function']['name'],
        "description": tool['function']['description'],
        "input_schema": tool['function']['parameters'],
    }
    for tool in DefinedToolsArray
]

Anthropic_Models = {
  "claude-3-7-sonnet-latest": {"company": "Anthropic","model": "claude-3-7-sonnet-latest", "input": 0.000003,    "output": 0.000015,   "context": 8192},
  "claude-3-5-sonnet-latest": {"company": "Anthropic","model": "claude-3-5-sonnet-latest", "input": 0.000003,    "output": 0.000015,   "context": 8192},
  "claude-3-5-haiku-latest":  {"company": "Anthropic","model": "claude-3-5-haiku-latest" , "input": 0.00000015,  "output": 0.000004,   "context": 8192},
}

AiRegistry.register_handler(company_name="Anthropic", handler_class=AiAnthropic)
AiRegistry.register_models_from_dict(model_definitions=Anthropic_Models)

